/**
 * Provides the basic interface for unique, serializable objects.
 */
import {array} from "prop-types";

export const GLOBAL_PARAMS = ["id"];

export class SerializableItem {
    public id: number;  // Serialized objects need to be unique.

    /**
     * Returns the current object in its serialized form.
     * @returns The serialized object.
     */
    public serialize(): any {
        throw new Error("Method 'serialize()' must be implemented.");
    }

    /**
     * Creates the current object based on raw, serialized data.
     * @param raw The serialized data.
     * @returns Parsed data in final form. Could be the finalisd object.
     */
    public static parse(raw: any): any {
        throw new Error("Method 'parse()' must be implemented.");
    }

    /**
     * A generic way to create a new object with all the desired parameters of the current one.
     * @param params List of parameters to include in the new object.
     * @param data The data object to be serialized. The current object by default.
     * @protected
     * @returns New object containing all the desired properties.
     */
    protected serializeProperties(params: string[], data: any = this): any {
        const serialized: any = {};

        params.forEach((param) => {
            serialized[param] = this.serializeItem((this as any)[param]);
        });

        return serialized;
    }

    /**
     * Recursively serializes an object. Handles serializable items and lists properly.
     * @param value Object to be serialized.
     * @private
     * @returns Serialized item.
     */
    private serializeItem(value: any): any {
        if (value instanceof SerializableItem) {
            // If is also serializable, use the serialized form
            return value.serialize();
        } else if (value instanceof array) {
            // If is some kind of list, convert the objects in the list recursively
            const serializedList: any = [];
            (value as []).forEach((item) => {
                serializedList.push(this.serializeItem(item));
            })
            return serializedList;
        } else {
            return value;
        }
    }
}
